# frozen_string_literal: true

require_relative '../response'
require_relative '../util/multipart'
require_relative '../util/header'

module Hyde
  module DSL
    # Common methods for Probe objects
    module ProbeMethods
      # Get the current request
      # @return [Hyde::Request]
      def request
        @origin.request
      end

      # Set response status (generate response if one doesn't exist yet)
      # @param status [Integer] http status code
      def status(status)
        @origin.response = (@origin.response or Hyde::Response.new)
        @origin.response.status = status
      end

      alias code status

      # Set response header (generate response if one doesn't exist yet)
      # @param key [String] header name
      # @param value [String] header value
      def header(key, value)
        return status(value) if key.downcase == "status"

        if key.match(/(?:[(),\/:;<=>?@\[\]{}"]|[^ -~])/)
          raise StandardError, "header key has invalid characters"
        end

        if value.match(/[^ -~]/)
          raise StandardError, "value key has invalid characters"
        end

        @origin.response = (@origin.response or Hyde::Response.new)
        key = key.downcase
        @origin.response.add_header(key, value)
      end

      # Delete a header value from the headers hash
      # If no value is provided, deletes all key entries
      # @param key [String] header name
      # @param value [String, nil] header value
      def remove_header(key, value = nil)
        return unless @origin.response

        return if key.downcase == "status"

        if key.match(/(?:[(),\/:;<=>?@\[\]{}"]|[^ -~])/)
          raise StandardError, "header key has invalid characters"
        end

        if value&.match(/[^ -~]/)
          raise StandardError, "value key has invalid characters"
        end

        @origin.response.delete_header(key, value)
      end

      # Checks if current request has multipart/form-data associated with it
      # @return [Boolean]
      def form?
        value, opts = Hyde::Util.parse_value(request.headers["content-type"])
        if value == "multipart/form-data" and
           opts["boundary"]
          true
        else
          false
        end
      end

      # Returns formdata
      # @return [Hash{String=>(String,Hyde::Util::FormPart)}]
      def form
        _, opts = Hyde::Util.parse_value(request.headers["content-type"])
        Hyde::Util::MultipartParser.new(
          request.input, opts["boundary"]
        ).to_h
      end
    end
  end
end
