# frozen_string_literal: true

require_relative 'pattern_matching'
require_relative 'node'
require_relative 'dsl/path_constructors'
require_relative 'dsl/path_methods'
require_relative 'dsl/common_methods'
require_relative 'util/lookup'

module Hyde
  # Binding that provides execution context for filters and preprocessors.
  class ProcessorBinding
    include Hyde::DSL::CommonMethods

    def initialize(path)
      @origin = path
    end
  end

  # Binding that provides execution context for path setup block.
  class PathBinding
    include Hyde::DSL::PathConstructors
    include Hyde::DSL::PathMethods

    def initialize(path)
      @origin = path
    end
  end

  # Primary building block of request navigation.
  class Path < Hyde::Node
    ProcBinding = Hyde::ProcessorBinding
    Binding = Hyde::PathBinding

    # @param path [Object] Object to generate {Hyde::Pattern} from
    # @param parent [Hyde::Node] Parent object to inherit properties to
    # @param setup [#call] Setup block
    def initialize(path, parent:, &setup)
      super(path, parent: parent)
      # Child nodes array
      @children = []
      # Arrays of preprocessors, postprocessors and filters
      @preprocessors = []
      @postprocessors = []
      @filters = []
      # Bindings setup
      binding = self.class::Binding.new(self)
      binding.instance_exec(&setup)
      @procbinding = self.class::ProcBinding.new(self)
    end

    # Method callback on successful request navigation.
    # Finds the next appropriate path to go to.
    # @return [Boolean] true if further navigation will be done
    # @raise [UncaughtThrowError] by default throws :response if no matches found.
    def process(request)
      return false unless run_filters(request)

      run_preprocessors(request)
      enqueue_postprocessors(request)
      @children.each do |x|
        if (value = x.go(request))
          return value
        end
      end
      value = index(request)
      return value if value

      _die(404)
    rescue StandardError => e
      _die(500, backtrace: [e.to_s] + e.backtrace)
    end

    # Add a preprocessor to the path.
    # Does not modify path execution.
    # @param block [#call]
    # @yieldparam request [Hyde::Request]
    def preprocess(&block)
      @preprocessors.append(block)
    end

    # Add a postprocessor to the path.
    # @param block [#call]
    # @yieldparam request [Hyde::Request]
    # @yieldparam response [Hyde::Response]
    def postprocess(&block)
      @postprocessors.append(block)
    end

    # Add a filter to the path.
    # Blocks path access if a filter returns false.
    # @param block [#call]
    # @yieldparam request [Hyde::Request]
    def filter(&block)
      @filters.append(block)
    end

    attr_reader :children, :properties

    private

    # Sequentially run through all filters and drop request if one is false
    # @param request [Hyde::Request]
    # @return [Boolean] true if request passed all filters
    def run_filters(request)
      @filters.each do |filter|
        return false unless @procbinding.instance_exec(request, &filter)
      end
      true
    end

    # Sequentially run all preprocessors on a request
    # @param request [Hyde::Request]
    def run_preprocessors(request)
      @preprocessors.each do |preproc|
        @procbinding.instance_exec(request, &preproc)
      end
    end

    # Append postprocessors to request
    # @param request [Hyde::Request]
    def enqueue_postprocessors(request)
      request.postprocessors.append(*@postprocessors)
    end

    # Try to perform indexing on the path if possible
    # @param request [Hyde::Request]
    # @return [Boolean] true if indexing succeeded
    def index(request)
      return false unless request.path.match?(/^\/?$/)

      @properties["index"].each do |index|
        request.path = index
        @children.each do |x|
          if (value = x.go(request))
            return value
          end
        end
      end
      false
    end

    # Handle an errorcode
    # @param errorcode [Integer]
    # @param backtrace [Array(String), nil]
    # @raise [UncaughtThrowError] throws :finish to stop processing
    def _die(errorcode, backtrace: nil)
      throw :finish, [errorcode].append(
        *(@properties["handle.#{errorcode}"] or
          @properties["handle.default"]).call(
            errorcode,
            backtrace: backtrace
          )
      )
    end
  end
end
