# frozen_string_literal: true

require_relative 'pattern_matching/util'
require_relative 'pattern_matching/glob'
require_relative 'pattern_matching/rematch'

module Hyde
  # Utility functions and pattern-generator classes.
  # Used primarily to create patterns for path definitions.
  module PatternMatching end

  # Delegate class for all available patterns.
  # Picks appropriate pattern based on contents.
  class Pattern
    # @param pattern [Object] object to generate pattern from
    def initialize(pattern)
      @pattern = patternify(pattern)
      @static = @pattern.is_a? String
    end

    # Checks if pattern object is static (is a simple String pattern).
    def static?
      @static
    end

    # Match the string and assign matches to parameters.
    # Returns:
    #
    # - Unmatched part of a string
    # - Unnamed parameters
    # - Named parameters
    # @param input [String] String to match
    # @return [Array(String,Array,Hash)]
    # @return [FalseClass] if input doesn't match pattern
    def match(input)
      if @pattern.is_a? String
        input = Hyde::PatternMatching.canonicalize(input)
        if input.start_with?(@pattern)
          [input.delete_prefix(@pattern), [], {}]
        else
          false
        end
      else
        @pattern.match(input)
      end
    end

    # Test if a string can be matched.
    # Lighter version of match that doesn't assign any variables.
    # @param input [String] String to match
    # @return [Boolean]
    def match?(input)
      if @pattern.is_a? String
        Hyde::PatternMatching.canonicalize(input).start_with? @pattern
      else
        @pattern.match?(input)
      end
    end

    private

    def patternify(pattern)
      classdomain = Hyde::PatternMatching
      classdomain.constants
                 .filter { |x| classdomain.const_get(x).is_a? Class }
                 .map { |x| classdomain.const_get(x) }
                 .each do |pattern_generator|
        if pattern_generator.can_convert? pattern
          return pattern_generator.new(pattern)
        end
      end
      Hyde::PatternMatching.canonicalize(pattern)
    end
  end
end
