# frozen_string_literal: true

module Hyde
  module PatternMatching
    # someday i've heard this quote:
    #   "if you love regexp so much why don't you marry the damn thing?"
    # and so i did, as you can see here.

    # Implements glob-like pattern matching.
    # Exact specifications for globbing rules:
    #
    # "/"
    #   - act as directory separators
    #   - multiple slashes (i.e. "///") are the same as one slash ("/")
    #   - slashes are stripped at start and end of an expression or path
    #   - slashes are not matched by anything but the globstar ("**")
    #
    # "*" ( regexp: /([^/]*)/ )
    #   - matches from 0 to any number of characters
    #   - does not match nothing if placed between two slashes (i.e "/*/")
    #   - result is captured in an array
    #   - stops at slashes
    #   - greedy (matches as much as possible)
    #
    # "**" ( regexp: /(.*)/ )
    #   - matches any number of characters
    #   - matches slashes ("/")
    #   - result is captured in an array
    #   - does not stop at slashes
    #   - greedy (matches as much as possible)
    #
    # "[...]+" ( regexp: itself, ! and ^ at the start are interchangeable )
    #   - acts like a regexp range
    #   - matches any characters, including slashes if specified
    #   - matches any number of characters
    #   - valid ways to specify a range: [A-z], [a-z], [9-z] (ascii order)
    #   - ! or ^ at the start invert meaning (any character not in range)
    #   - result is captured in an array
    #
    # ":name" ( regexp: acts like a named group for /[^/]*/ )
    #   - acts like * as defined above
    #   - result is captured in a hash with "name" as key
    #   - name allows alphanumeric characters and underscores
    #   - ignored unless placed between two slashes
    #
    # "(name|name2|...)" ( regexp: itself )
    #   - matches any of (full keywords) name, name2, ...
    #   - result is captured in an array
    #   - each name may include only these groups of characters:
    #     - alphanumeric
    #     - slashes
    #     - underscores
    #     - dashes
    class Glob
      # @param input [String] Glob pattern
      def initialize(pattern)
        pattern = Hyde::PatternMatching.canonicalize(pattern)
        pieces = pattern.split(TOKENS)
        # @type [Array<String,Integer>]
        @index = build_index(pieces)
        # @type [Regexp]
        @glob = build_regexp(pieces)
      end

      # Match the string and assign matches to parameters.
      # Returns:
      #
      # - Unmatched part of a string
      # - Unnamed parameters
      # - Named parameters
      # @param input [String] String to match
      # @return [Array(String,Array,Hash)]
      # @return [FalseClass] if input doesn't match glob
      def match(input)
        input = Hyde::PatternMatching.canonicalize(input)
        result = input.match(@glob)
        return false unless result

        input = result.post_match
        named_params, params = assign_by_index(@index, result.captures)
        [input, params, named_params]
      end

      # Test if a string can be matched.
      # Lighter version of match that doesn't assign any variables.
      # @param input [String] String to match
      # @return [Boolean]
      def match?(input)
        input = Hyde::PatternMatching.canonicalize(input)
        input.match? @glob
      end

      # Test if input is convertible to a Glob and if it should be converted
      # @param input
      # @return [Boolean] Input can be safely converted to Glob
      def self.can_convert?(input)
        input.is_a? String and
          input.match?(TOKENS)
      end

      private

      # i shall name thee Norxondor Glorbulon

      # Regexp pattern to match glob tokens
      TOKENS = /
        ( # Glob-specific tokens
          (?<=(?:\/|^))\*\*(?:\/|$) |     # Freestanding globstar
          \* |                            # Regular glob
          \[!?\w-\w\]\+ |                 # Character group
          (?<=\/):[\w_]+(?=(?:\/|$)) |    # Named glob
          \([\w\/|_-]+\)                  # Alternator group
        )
      /x.freeze

      # Build an index for separating normal matches from named matches
      # @param pieces [Array(String)] Glob pattern after being splitted
      # @return [Array(String,Integer)] Index array for assign_by_index
      def build_index(pieces)
        count = -1
        index = []
        pieces.each do |x|
          if (name = x.match(/^:([\w_]+)$/))
            index.append(name[1])
          elsif x.match?(TOKENS)
            index.append(count += 1)
          end
        end
        index
      end

      # Assign values from match.captures to named and numbered groups
      # @param index [Array(String,Integer)] Index array
      # @param params [Array] Unnamed captures from a String.match
      # @return [Array(Hash,Array)]
      def assign_by_index(index, params)
        named_params = {}
        new_params = []
        params.each_with_index do |x, k|
          # FIXME: this is a hack for globstar implementation at line 152
          x = (x.nil? ? "" : x)
          if index[k].is_a? String
            named_params[index[k].to_sym] = x
          else
            new_params[index[k]] = x
          end
        end
        [named_params, new_params]
      end

      # Create a Regexp from a glob pattern according to defined syntax
      # @param tokens [Array(String)] Split array of glob tokens
      # @return [Regexp]
      def build_regexp(tokens)
        Regexp.new(tokens.map do |filter|
          case filter
          when /(\/|^)\*\*(\/|$)/ then "(?:(.*)/|)" # FIXME: this may return nil
          when "*" then "([^/]*)"
          when /^\([\w\/|_-]+\)$/ then filter.sub('-', '\\-')
          when /^\[!?\w-\w\]\+$/ then filter.sub('!', '^')
          when /^:[\w_]+$/ then "([^/]*)"
          else filter.gsub(/[\^$.|+(){}]/, '\\\\\\0')
          end
        end.join("").prepend("^/?"))
      end
    end
  end
end
