# frozen_string_literal: true

module Hyde
  module PatternMatching
    # Regexp pattern matching wrapper.
    # Following principles apply to the wrapper:
    #
    # - Regexp input is canonicalized using Hyde::PatternMatching.canonicalize
    # - Matched content and anything before it is consumed in #match output
    #
    # @note If you are planning to write your own pattern, this is the easiest one to read.
    class ReMatch
      def initialize(pattern)
        @glob = pattern
      end

      # Match the string and assign matches to parameters.
      # Returns:
      #
      # - Unmatched part of a string
      # - Unnamed parameters
      # - Named parameters
      # @param input [String] String to match
      # @return [Array(String,Array,Hash)]
      # @return [FalseClass] if input doesn't match regexp
      def match(input)
        result = Hyde::PatternMatching.canonicalize(input).match(@glob)
        return false unless result

        [result.post_match, result.captures, result.named_captures]
      end

      # Test if a string can be matched.
      # Lighter version of match that doesn't assign any variables.
      # @param input [String] String to match
      # @return [Boolean]
      def match?(input)
        Hyde::PatternMatching.canonicalize(input).match? @glob
      end

      # Test if input is convertible and if it should be converted.
      # @param input
      # @return [Boolean] Input can be safely converted to Glob
      def self.can_convert?(string)
        string.is_a? Regexp
      end
    end
  end
end
