# frozen_string_literal: true

module Hyde
  # Probe that executes a callback on request
  class Handler < Hyde::Probe
    # @param path [Object]
    # @param parent [Hyde::Node]
    # @param exec [#call]
    def initialize(path, parent:, &exec)
      super(path, parent: parent)
      @callback = exec
      @binding = Hyde::ProbeBinding.new(self)
      @response = nil
    end

    attr_accessor :response
    attr_reader :request

    # Method callback on successful request navigation.
    # Runs block supplied with object initialization.
    # Request's #splat and #param are passed to block.
    #
    # Callback's returned should be one of viable responses:
    #
    # - {Hyde::Response} object
    # - An array that matches Rack return form
    # - An array that matches old (Rack 2.x) return form
    # - A string (returned as HTML with code 200)
    # - false (bounces the request to next handler)
    # @param request [Hyde::Request]
    # @return [Boolean] true if further navigation is possible
    # @raise [UncaughtThrowError] may raise if die() is called.
    def process(request)
      @response = nil
      return reject(request) unless request.path.match?(/^\/?$/)

      @request = request
      response = catch(:break) do
        @binding.instance_exec(*request.splat,
                               **request.param,
                               &@callback)
      end
      return false unless response

      if @response and [String, File, IO].include? response.class
        @response.body = response
        throw :finish, @response
      end
      throw :finish, response
    end
  end
end
