# frozen_string_literal: true

module Hyde
  # Rack protocol response wrapper.
  class Response
    @chunk_size = 1024

    self.class.attr_accessor :chunk_size

    # @param response [Array(Integer, Hash, Array), nil]
    def initialize(response = nil)
      if response
        @status = response[0]
        @headers = response[1]
        @body = response[2]
      else
        @status = 404
        @headers = {}
        @body = []
      end
    end

    # Return internal representation of Rack response
    # @return [Array(Integer,Hash,Array)]
    def finalize
      [@status, @headers, @body]
    end

    # Make internal representation conformant
    # @return [Hyde::Response]
    def validate
      if [204, 304].include?(@status) or (100..199).include?(@status)
        @headers.delete "content-length"
        @headers.delete "content-type"
        @body = []
      elsif @headers.empty?
        @headers = {
          "content-length" => content_size,
          "content-type" => "text/html"
        }
      end
      @body = self.class.chunk_body(@body) if @body.is_a? String
      self
    end

    # Add a header to the headers hash
    # @param key [String] header name
    # @param value [String] header value
    def add_header(key, value)
      if @headers[key].is_a? String
        @headers[key] = [@headers[key], value]
      elsif @headers[key].is_a? Array
        @headers[key].append(value)
      else
        @headers[key] = value
      end
    end

    # Delete a header value from the headers hash
    # If no value is provided, deletes all key entries
    # @param key [String] header name
    # @param value [String, nil] header value
    def delete_header(key, value = nil)
      if value and @response[key]
        @response[key].delete(value)
      else
        @response.delete(key)
      end
    end

    attr_accessor :status, :headers, :body

    # Ensure response correctness
    # @param obj [String, Array, Hyde::Response]
    # @return Response
    def self.convert(obj)
      case obj
      when Response
        obj.validate
      when Array
        Response.new(obj).validate
      when String, File, IO
        Response.new([200,
                      {
                        "content-type" => "text/html"
                      },
                      obj]).validate
      else
        Response.new([404, {}, []])
      end
    end

    # Turn body into array of chunks
    # @param text [String]
    # @return [Array(String)]
    def self.chunk_body(text)
      text.chars.each_slice(@chunk_size).map(&:join)
    end

    private

    # Try to figure out content length
    # @return [Integer, nil]
    def content_size
      case @body
      when String then @body.length
      when Array then @body.join.length
      when File then @body.size
      end
    end
  end
end
