# frozen_string_literal: true

require 'uri'
require_relative 'sorting'

module Hyde
  module Util
    # TODO: encoding support????
    # Query string parser
    class Query
      include Hyde::Util::ParserCommon
      # @param query [String]
      def initialize(query)
        @query = query
      end

      # Shallow query parser (does not do PHP-like array keys)
      # @return [Hash]
      def parse_shallow
        URI.decode_www_form(@query)
           .sort_by { |array| array[0] }
           .to_h
      end

      # Better(tm) query parser.
      # Returns a hash with arrays.
      # Key semantics:
      #
      # - `key=value` creates a key value pair
      # - `key[]=value` appends `value` to an array named `key`
      # - `key[index]=value` sets `value` at `index` of array named `key`
      # @return [Hash]
      def parse
        construct_deep_hash(URI.decode_www_form(@query))
      end

      # Get key from query.
      # @param key [String]
      # @return [String,Array]
      def [](key)
        (@cache ||= parse)[key]
      end

      private

      # Construct a hash with array support
      def construct_deep_hash(array)
        flatten(sort(group(array)))
      end

      # Assign values to keys in a new hash and group arrayable keys
      def group(array)
        hash = {}
        array.each do |key, value|
          if key.match?(/.*\[\d*\]\Z/)
            new_key, index = key.match(/(.*)\[(\d*)\]\Z/).to_a[1..]
            hash[new_key] = [] unless hash[new_key]
            hash[new_key].append([index, value])
          else
            hash[key] = value
          end
        end
        hash
      end
    end
  end
end
